unit mleTagResolvers;

interface

uses
  Classes, DBTables, dpoBase, mleCommon, mlePropertyInfo, mleSMLContainer, usXMLDoc;

type
  TControlType = (ctUnknown, ctEditBox, ctCheckBox, ctSelect, ctTable);

const
  CONTROL_TYPE_NAMES: array[TControlType] of string =
    ('', 'edit', 'checkbox', 'select', 'table');

type
  TTagResolver = class;
  TTagResolverClass = class of TTagResolver;

  TTagResolver = class
  private
    FContainer: TSMLContainer;
    FParent: TTagResolver;
    FTagElement: TusXMLElement;
  protected
    HTML: TStringList;

    { Functional methods }
    function EvaluateExpression(aExpression: string): string; virtual;
    function ExpandText(aText: string): string; virtual;
    procedure GetAttributes; virtual;
    procedure GetSubtags;
    procedure HandleSubtag(aElement: TusXMLElement; var aHandled: Boolean); virtual;

    { Helper methods }
    function GetAttribute(aName, aDefault: string): string;
  public
    constructor Create(aContainer: TSMLContainer; aParent: TTagResolver; aTagElement: TusXMLElement);
    destructor Destroy; override;
    function GetHTML: string; virtual;
    procedure Resolve; virtual;
    procedure Setup; virtual;
  end;

  TSimpleControlTagResolver = class(TTagResolver)
  protected
    AType: string;
    AName: string;
    AValue: string;
    procedure GetAttributes; override;
  public
    procedure Resolve; override;
  end;

  TSMLTagResolver = class(TTagResolver)
  protected
    procedure HandleSubtag(aElement: TusXMLElement; var aHandled: Boolean); override;
  end;

  TDataObjectTagResolver = class(TTagResolver)
  protected
    AClassName: string;
    AName: string;
    AOID: string;
    AClass: TDataObjectClass;
    procedure GetAttributes; override;
  public
    procedure Resolve; override;
  end;

  TControlTagResolver = class(TTagResolver)
  protected
    AType: string;
    AName: string;
    APropertyName: string;
    ASize: string;
    AMaxLength: string;
    AValue: string;
    ACols, ARows: string;
    XType: TControlType;
    PropertyInfo: TPropertyInfo;
    procedure GetAttributes; override;
    function GetHTMLAttributes: string; virtual;
  public
    destructor Destroy; override;
    procedure Resolve; override;
  end;

  TTextTagResolver = class(TTagResolver)
  public
    function GetHTML: string; override;
  end;

  TLinkTagResolver = class(TTagResolver)
  protected
    APage: string;
    HRef: string;
    Params: string;
    procedure GetAttributes; override;
    procedure HandleSubtag(aElement: TusXMLElement; var aHandled: Boolean); override;
  public
    function GetHTML: string; override;
  end;

  TButtonTagResolver = class(TTagResolver)
  protected
    AType: string;
    ACaption: string;
    XType: string;
    URLParams: string;
    procedure GetAttributes; override;
    function GetOnClick: string;
    procedure HandleSubtag(aElement: TusXMLElement; var aHandled: Boolean); override;
  public
    function GetHTML: string; override;
  end;

  TActionTagResolver = class(TTagResolver)
  protected
    procedure GetAttributes; override;
    procedure HandleSubtag(aElement: TusXMLElement; var aHandled: Boolean); override;
  public
    AType: string;
    AValue: string;
    Params: string;
    procedure Resolve; override;
  end;

  TParamTagResolver = class(TTagResolver)
  protected
  public
    procedure Resolve; override;
  end;

implementation

uses
  DB, SysUtils;

const
  ROUTER_EXE = '../cgi-bin/router.exe';

{ TTagResolver }

constructor TTagResolver.Create(aContainer: TSMLContainer; aParent: TTagResolver; aTagElement: TusXMLElement);
begin
  inherited Create;
  FContainer := aContainer;
  FParent := aParent;
  FTagElement := aTagElement;
  HTML := TStringList.Create;
end;

destructor TTagResolver.Destroy;
begin
  HTML.Free;
  inherited;
end;

function TTagResolver.EvaluateExpression(aExpression: string): string;
begin
  Result := Trim(aExpression);
  if Copy(Result, 1, 2) = '${' then
  begin
    aExpression := Copy(Result, 3, Length(Result) - 3);
    if FContainer.Variables.IndexOfName(aExpression) <> -1 then
      Result := FContainer.Variables.Values[aExpression]
    else
      if Pos('.', aExpression) <> 0 then
        Result := FContainer.ObjectCache.GetPropertyValue(aExpression)
      else
        raise Exception.CreateFmt('Unknown expression: %s', [aExpression]);
  end;
end;

function TTagResolver.ExpandText(aText: string): string;
var
  P, I, L: Integer;
  Viewport: string;
  ExpressionResult: string;
begin
  Result := aText;

  { Collapse paired quote characters and expand embedded expressions }
  P := 1;
  while P < Length(Result) do
  begin
    Viewport := Copy(Result, P, 2);
    if Viewport = #39#39 then     { Collapse paired quote characters }
      Delete(Result, P, 1)
    else if Viewport = '${' then  { Expand embedded expression }
    begin
      I := P + 2;
      L := Length(Result);
      while (I <= L) and (Result[I] <> '}') do
        Inc(I);
      if I > L then
        raise Exception.Create('Unterminated expression: ' + aText);  {!!}
      ExpressionResult := EvaluateExpression(Copy(Result, P, I - P + 1));
      Delete(Result, P, I - P + 1);
      Insert(ExpressionResult, Result, P);
      Inc(P, Length(ExpressionResult) - 1);
    end;
    Inc(P);
  end;
end;

function TTagResolver.GetAttribute(aName, aDefault: string): string;
var
  Attr: TusXMLAttribute;
begin
  Result := aDefault;
  Attr := FTagElement.Attributes.GetByName(aName);
  if Assigned(Attr) then
    Result := Attr.Value;
end;

procedure TTagResolver.GetAttributes;
begin
  // do nothing; not all tags will have attributes
end;

function TTagResolver.GetHTML: string;
begin
  Result := '';
  if HTML.Count <> 0 then
    Result := HTML.Text;
end;

procedure TTagResolver.GetSubtags;
var
  I: Integer;
  Handled: Boolean;
begin
  for I := 0 to FTagElement.Subtags.Count - 1 do
  begin
    Handled := False;
    HandleSubtag(FTagElement.Subtags[I], Handled);
    if not Handled then
      raise Exception.CreateFmt('Unexpected subtag <%s>', [FTagElement.Subtags[I].TagName]);
  end;
end;

procedure TTagResolver.HandleSubtag(aElement: TusXMLElement;
  var aHandled: Boolean);
begin
  // Do nothing; not all tags will have subtags
end;

procedure TTagResolver.Resolve;
begin
  // Do nothing; logic is tag-specific
end;

procedure TTagResolver.Setup;
begin
  GetAttributes;
  GetSubtags;
end;

{ TSMLTagResolver }

procedure TSMLTagResolver.HandleSubtag(aElement: TusXMLElement;
  var aHandled: Boolean);
const
  NUM_ENTRIES = 5;
  TagTable: array[0..NUM_ENTRIES - 1] of record
    TagName: string;
    ResolverClass: TTagResolverClass;
  end = ((TagName:'BUTTON'; ResolverClass:TButtonTagResolver),
         (TagName:'CONTROL'; ResolverClass:TControlTagResolver),
         (TagName:'DATAOBJECT'; ResolverClass:TDataObjectTagResolver),
         (TagName:'LINK'; ResolverClass:TLinkTagResolver),
         (TagName:'TEXT'; ResolverClass:TTextTagResolver)
         );
var
  I: Integer;
  S: string;
begin
  for I := 0 to NUM_ENTRIES - 1 do
    with TagTable[I] do
      if aElement.TagName = TagName then
      begin
        with ResolverClass.Create(FContainer, Self, aElement) do
          try
            Setup;
            Resolve;
            S := GetHTML;
            if S <> '' then
              Self.HTML.Add(S);
          finally
            Free;
          end;
        aHandled := True;
        Break;
      end;
end;

{ TControlTagResolver }

destructor TControlTagResolver.Destroy;
begin
  PropertyInfo.Free;
  inherited;
end;

procedure TControlTagResolver.GetAttributes;
begin
  inherited;
  AType := GetAttribute('type', '');
  AName := GetAttribute('name', '');
  APropertyName := GetAttribute('property', '');
  PropertyInfo := TPropertyInfo.Create;
  FContainer.ObjectCache.GetPropertyInfo(APropertyName, PropertyInfo);
  AValue := GetAttribute('value', '');
  ASize := GetAttribute('size', '');
  AMaxLength := GetAttribute('maxlength', '');
  ACols := GetAttribute('cols', '');
  ARows := GetAttribute('rows', '');
end;

function TControlTagResolver.GetHTMLAttributes: string;
begin
  Result := '';
  if AName <> '' then
    Result := Result + ' name=' + AName;
  if ASize <> '' then
    Result := Result + ' size=' + ASize;
  if AMaxLength <> '' then
    Result := Result + ' maxlength=' + AMaxLength;  

  if AType = 'memo' then
  begin
    if ACols <> '' then
      Result := Result + ' cols=' + ACols;
    if ARows <> '' then
      Result := Result + ' rows=' + ARows;
  end;
end;

procedure TControlTagResolver.Resolve;
begin
  if APropertyName <> '' then
  begin
    AValue := PropertyInfo.AsString;

    if ASize = '' then
      if PropertyInfo.DataType in [ftString] then
        ASize := IntToStr(PropertyInfo.Size);

    if AMaxLength = '' then
      if PropertyInfo.DataType in [ftString, ftMemo] then
        AMaxLength := IntToStr(PropertyInfo.Size);
  end;

  { If we've bothered to assign a name to the control, assume it's editable }
  if AName <> '' then
    FContainer.DataBindings.Add(PropertyInfo, AName);

  if AType = 'edit' then
  begin
    HTML.Add(Format('<INPUT type="text" value="%s"%s>', [AValue, GetHTMLAttributes]));
  end
  else if AType = 'memo' then
  begin
    HTML.Add(Format('<TEXTAREA%s>%s</TEXTAREA>', [GetHTMLAttributes, AValue]));
  end;
end;

{ TSimpleControlTagResolver }

procedure TSimpleControlTagResolver.GetAttributes;
begin
  AType := GetAttribute('type', '');
  if AType <> 'edit' then
    raise Exception.CreateFmt('Illegal value for "type" attribute: %s', [AType]);
  AName := GetAttribute('name', '');
  AValue := GetAttribute('value', '');
end;

procedure TSimpleControlTagResolver.Resolve;
begin
  if AType = 'edit' then
  begin
    HTML.Add(Format('<INPUT type="text" name="%s" value="%s">',
                    [AName, aValue]));
  end;
end;

{ TDataObjectTagResolver }

procedure TDataObjectTagResolver.GetAttributes;
var
  TempClass: TPersistentClass;
begin
  AClassName := GetAttribute('class', '');
  AName := GetAttribute('name', '');
  AOID := EvaluateExpression(GetAttribute('oid', ''));

  TempClass := GetClass(AClassName);
  if not Assigned(TempClass) or not TempClass.InheritsFrom(TDataObject) then
    raise Exception.CreateFmt('%s is not a valid class', [AClassName]);
  AClass := TDataObjectClass(TempClass);
end;

procedure TDataObjectTagResolver.Resolve;
var
  Instance: TDataObject;
begin
  Instance := AClass.Create(FContainer.Database);
  Instance.Name := AName;
  FContainer.ObjectCache.AddObject(AName, Instance);
  if AOID <> '' then
    Instance.GetByOID(AOID);
end;

{ TTextTagResolver }

function TTextTagResolver.GetHTML: string;
begin
  Result := ExpandText(FTagElement.Data);
end;

{ TLinkTagResolver }

procedure TLinkTagResolver.GetAttributes;
begin
  APage := GetAttribute('page', '');
  HRef := Format('%s?PI:PAGE=%s', [ROUTER_EXE, APage]);
end;

function TLinkTagResolver.GetHTML: string;
begin
  Result := Format('<A href="%s%s">%s</A>', [HRef, Params, FTagElement.Data]);
end;

procedure TLinkTagResolver.HandleSubtag(aElement: TusXMLElement;
  var aHandled: Boolean);
begin
  if aElement.TagName = 'PARAM' then
  begin
    with TParamTagResolver.Create(FContainer, Self, aElement) do
      try
        Setup;
        Params := Params + '&' + ExpandText(aElement.Data);
      finally
        Free;
      end;
    aHandled := True;
  end;
end;

{ TButtonTagResolver }

procedure TButtonTagResolver.GetAttributes;
begin
  AType := GetAttribute('type', 'button');
  XType := AType;
  if CompareText(XType, 'submit') = 0 then
    XType := 'button';
  ACaption := GetAttribute('caption', '');
end;

function TButtonTagResolver.GetHTML: string;
begin
  Result := Format('<INPUT type=%s value=%s onclick=%s>', [XType, ACaption, GetOnClick]);
end;

function TButtonTagResolver.GetOnClick: string;
begin
  Result := '';
  if CompareText(AType, 'submit') = 0 then
  begin
    with TStringList.Create do
      try
        Add('function DoSubmit(params) {');
        Add('  document.forms[0].method=''post''');
        Add('  document.forms[0].action=''../cgi-bin/router.exe?'' + params');
        Add('  document.forms[0].submit()');
        Add('}');
        FContainer.Scripts.Values['DoSubmit'] := Text;
        Result := Format('DoSubmit("%s")', [URLParams]);
      finally
        Free;
      end;
  end;
end;

procedure TButtonTagResolver.HandleSubtag(aElement: TusXMLElement;
  var aHandled: Boolean);
begin
  if aElement.TagName = 'ACTION' then
  begin
    with TActionTagResolver.Create(FContainer, Self, aElement) do
      try
        Setup;
        if URLParams <> '' then
          URLParams := URLParams + '&';
        URLParams := URLParams + Format('PI:%s=%s', [AType, AValue]);
        if Params <> '' then
          URLParams := URLParams + '&' + Params;
      finally
        Free;
      end;
    aHandled := True;
  end;
end;

{ TActionTagResolver }

procedure TActionTagResolver.GetAttributes;
begin
  AType := GetAttribute('type', '');
  AValue := GetAttribute('value', '');
end;

procedure TActionTagResolver.HandleSubtag(aElement: TusXMLElement;
  var aHandled: Boolean);
begin
  if aElement.TagName = 'PARAM' then
  begin
    with TParamTagResolver.Create(FContainer, Self, aElement) do
      try
        Setup;
        if Params <> '' then
          Params := Params + '&';
        Params := Params + ExpandText(aElement.Data);
      finally
        Free;
      end;
    aHandled := True;
  end;
end;

procedure TActionTagResolver.Resolve;
begin

end;

{ TParamTagResolver }

procedure TParamTagResolver.Resolve;
begin

end;

end.
