unit RegValet;

interface

uses
  SysUtils, Classes, Windows, Registry, Forms;

type
  ERegValetError = class(Exception);

  TRegValetItem = class(TCollectionItem)
  private
    FName: String;
    FDefaultValue: String;
    FMaxOccurs: Integer;
  protected
    procedure SetDefaultValue(const Value: String);
    procedure SetName(const Name: String);
    procedure SetMaxOccurs(Value: Integer);
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
  published
    property DefaultValue: String read FDefaultValue
        write SetDefaultValue;
    property MaxOccurs: Integer read FMaxOccurs write SetMaxOccurs
        default 1;
    property Name: String read FName write SetName;
  end;

  TRegValetItems = class(TCollection)
  private
    FOwner: TPersistent;
  protected
    function GetItem(Index: Integer): TRegValetItem;
    procedure SetItem(Index: Integer; Item: TRegValetItem);
    function GetOwner: TPersistent; override;
  public
    constructor Create(Owner: TPersistent);
    function Add: TRegValetItem;
  public
    property Items[Index: Integer]: TRegValetItem read GetItem
        write SetItem; default;
  end;

  TRegValet = class(TComponent)
  private
    FEasyRegItems: TRegValetItems;
    FListDelim: String;
    FPath: String;
    FSuppressSave: Boolean;
    Names: TStringList;
    NonIndexedValues: TStringList;
    function GetCount(const Name: String): Integer;
    function GetIndexedListRef(const Name: String): TStringList;
    function GetIndexedValue(const Name: String; Index: Integer):
         String;
    function GetValue(const Name: String): String;
    function ItemIsIndexed(const Name: String): Boolean;
    procedure SetIndexedValue(const Name: String;
        Index: Integer; const Value: String);
    procedure SetListDelim(const Value: String);
    procedure SetValue(const Name: String; const Value: String);
    procedure StoreIndexedItem(const Name: String; Value: String);
    procedure StoreItemFromReg(const Name: String; const Value: String);
    procedure StoreNonIndexedItem(const Name: String;
        const Value: String);
  protected
    procedure Loaded; override;
    procedure MergeDefaults;
    procedure ReadRegistry;
    procedure WriteRegistry;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DeleteIndexedValue(const Name: String; Index: Integer);
        overload;
    procedure DeleteIndexedValue(const Name: String;
        const Value: String); overload;
    function IndexOf(const Name: String; const Value: String): Integer;
    procedure Initialize;
    procedure InsertIndexedValue(const Name: String;
        BeforeIndex: Integer; const Value: String);
    procedure MoveIndexedValueToFront(const Name: String;
        const Value: String); overload;
    procedure MoveIndexedValueToFront(const Name: String;
        Index: Integer); overload;
    procedure RestoreFormBounds;
    procedure SaveFormBounds;
  public
    property Count[const Name: String]: Integer read GetCount;
    property SuppressSave: Boolean read FSuppressSave
        write FSuppressSave;
    property IndexedValues[const Name: String; Index: Integer]: String
        read GetIndexedValue write SetIndexedValue;
    property Values[const Name: String]: String read GetValue
        write SetValue; default;
  published
    property Items: TRegValetItems read FEasyRegItems
        write FEasyRegItems;
    property Path: String read FPath write FPath;
    property ListDelim: String read FListDelim write SetListDelim;
  end;

procedure Register;

implementation

{ TRegValetItem }

procedure TRegValetItem.Assign(Source: TPersistent);
begin
  if Source is TRegValetItem then
  begin
    Name := TRegValetItem(Source).Name;
    DefaultValue := TRegValetItem(Source).DefaultValue;
    MaxOccurs := TRegValetItem(Source).MaxOccurs;
  end
  else
    inherited Assign(Source);
end;

constructor TRegValetItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FMaxOccurs := 1;
end;

procedure TRegValetItem.SetDefaultValue(const Value: String);
begin
  if (Value <> '') and (FMaxOccurs > 1) then
    raise ERegValetError.Create('Registry item with MaxOccurs > 1 ' +
        'may not be given an initial value');
    FDefaultValue := Value;
end;

procedure TRegValetItem.SetMaxOccurs(Value: Integer);
begin
  if Value < 1 then
    Value := 1;
  if (FDefaultValue <> '') and (Value > 1) then
    raise ERegValetError.Create('Registry item with a default value ' +
        'may not be given MaxOccurs > 1');
  FMaxOccurs := Value;
end;

procedure TRegValetItem.SetName(const Name: String);
begin
  if Name = '' then
    raise ERegValetError.Create('Registry item name may not be blank');
  FName := Name;
end;

{ TRegValetItems }

function TRegValetItems.Add: TRegValetItem;
begin
  Result := TRegValetItem(inherited Add);
end;

constructor TRegValetItems.Create(Owner: TPersistent);
begin
  inherited Create(TRegValetItem);
  FOwner := Owner;
end;

function TRegValetItems.GetItem(Index: Integer): TRegValetItem;
begin
  Result := TRegValetItem(inherited GetItem(Index));
end;

function TRegValetItems.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

procedure TRegValetItems.SetItem(Index: Integer; Item: TRegValetItem);
var
  I: Integer;
begin
  for I := 0 to (Count - 1) do
    if I <> Index then
      if Item.Name = Items[I].Name then
        raise ERegValetError.Create('Duplicate name in registry ' +
            'items: ' + Item.Name);
  inherited SetItem(Index, TCollectionItem(Item));
end;

{ TRegValet }

constructor TRegValet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEasyRegItems := TRegValetItems.Create(Self);
  FListDelim := ';';
  FPath := '\Software\ACompany\AnApp';
  Names := TStringList.Create;
  NonIndexedValues := TStringList.Create;
end;

procedure TRegValet.DeleteIndexedValue(const Name: String;
    Index: Integer);
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  if (Index >= 0) and (Index < L.Count) then
    L.Delete(Index);
end;

procedure TRegValet.DeleteIndexedValue(const Name: String;
    const Value: String);
var
  I: Integer;
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  I := L.IndexOf(Value);
  if I <> -1 then
    L.Delete(I);
end;

destructor TRegValet.Destroy;
var
  I: Integer;
begin
  if not FSuppressSave then
    WriteRegistry;
  FEasyRegItems.Free;
  for I := 0 to (Names.Count - 1) do
    if Names.Objects[I] <> nil then
      TStringlist(Names.Objects[I]).Free;
  Names.Free;
  NonIndexedValues.Free;
  inherited Destroy;
end;

function TRegValet.GetCount(const Name: String): Integer;
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  Result := L.Count;
end;

function TRegValet.GetIndexedListRef(const Name: String): TStringList;
var
  I: Integer;
begin
  I := Names.IndexOf(Name);
  if (I = -1) or (Names.Objects[I] = nil) then
    raise ERegValetError.Create('Indexed item name ' +
        Name + ' does not exist');
  Result := TStringList(Names.Objects[I]);
end;

function TRegValet.GetIndexedValue(const Name: String;
    Index: Integer): String;
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  if Index >= L.Count then
    raise ERegValetError.Create('Index out of bounds for ' +
        'GetIndexedValue call');
  Result := L[Index];
end;

function TRegValet.GetValue(const Name: String): String;
var
  I: Integer;
begin
  Result := '';
  I := Names.IndexOf(Name);
  if I <> -1 then
    Result := NonIndexedValues[I];
end;

function TRegValet.IndexOf(const Name: String; const Value: String):
    Integer;
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  Result := L.IndexOf(Value);
end;

procedure TRegValet.Initialize;
var
  I: Integer;
  ERI: TRegValetItem;
  L: TStringList;
begin
  for I := 0 to (FEasyRegItems.Count - 1) do
  begin
    ERI := FEasyRegItems[I];
    Names.Add(ERI.Name);
    NonIndexedValues.Add(ERI.DefaultValue);
    if ERI.MaxOccurs > 1 then
    begin
      L := TStringList.Create;
      L.Capacity := ERI.MaxOccurs;
      Names.Objects[I] := L;
    end;
  end;
end;

procedure TRegValet.InsertIndexedValue(const Name: String;
    BeforeIndex: Integer; const Value: String);
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  if L.Count = L.Capacity then
    L.Delete(L.Count - 1);
  if BeforeIndex >= L.Count then
    L.Add(Value)
  else
    L.Insert(BeforeIndex, Value);
end;

function TRegValet.ItemIsIndexed(const Name: String): Boolean;
var
  I: Integer;
  ERI: TRegValetItem;
begin
  Result := False;
  for I := 0 to (FEasyRegItems.Count - 1) do
  begin
    ERI := FEasyRegItems[I];
    if (ERI.Name = Name) and (ERI.MaxOccurs > 1) then
    begin
      Result := True;
      Exit;
    end;
  end;
end;

procedure TRegValet.Loaded;
begin
  inherited Loaded;
  Initialize;
  MergeDefaults;
  ReadRegistry;
end;

procedure TRegValet.MergeDefaults;
var
  I: Integer;
  ERI: TRegValetItem;
begin
  for I := 0 to (FEasyRegItems.Count - 1) do
  begin
    ERI := FEasyRegItems[I];
    if ERI.DefaultValue <> '' then
      Values[ERI.Name] := ERI.DefaultValue;
  end;
end;

procedure TRegValet.MoveIndexedValueToFront(const Name: String;
    const Value: String);
var
  I: Integer;
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  I := L.IndexOf(Value);
  if I <> -1 then
    L.Move(I, 0)
  else
    InsertIndexedValue(Name, 0, Value);
end;

procedure TRegValet.MoveIndexedValueToFront(const Name: String;
    Index: Integer);
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  if (Index > -1) and (Index < L.Count) then
    L.Move(Index, 0);
end;

procedure TRegValet.ReadRegistry;
var
  I: Integer;
  Reg: TRegistry;
  Names: TStringList;
begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if not Reg.OpenKeyReadOnly(FPath) then
      Exit;
    Names := TStringList.Create;
    try
      Reg.GetValueNames(Names);
      for I := 0 to (Names.Count - 1) do
        StoreItemFromReg(Names[I], Reg.ReadString(Names[I]));
    finally
      Names.Free;
    end;
    Reg.CloseKey;
  finally
    Reg.Free;
  end;
end;

procedure TRegValet.RestoreFormBounds;
var
  Left: Integer;
  Top: Integer;
  Height: Integer;
  Width: Integer;
  BoundsStr: String;

  function NextValue: Integer;
  var
    I: Integer;
  begin
    Result := -9999;
    I := Pos(';', BoundsStr);
    if I > 0 then
    begin
      Result := StrToInt(Copy(BoundsStr, 1, (I - 1)));
      BoundsStr := Copy(BoundsStr, (I + 1), $7FFF);
    end
    else
      BoundsStr := ''; // values been manually altered, give up
  end;

begin
  if (Owner is TCustomForm) then
  begin
    BoundsStr := Values['RegValetFormBounds'];
    if BoundsStr <> '' then
    begin
      try
        Left := NextValue;
        Top := NextValue;
        Width := NextValue;
        Height:= NextValue;
        if (Left <> -9999) and (Top <> -9999) and (Width <> -9999) and
            (Height <> -9999) then
          TCustomForm(Owner).SetBounds(Left, Top, Width, Height);
      except
        // don't make a big deal out of it
      end;
    end;
  end;
end;

procedure TRegValet.SaveFormBounds;
var
  Placement: TWindowPlacement;
  R: TRect;
begin
  if (Owner is TCustomForm) then
  begin
    Placement.Length := SizeOf(TWindowPlacement);
    GetWindowPlacement(TCustomForm(Owner).Handle, @Placement);
    R := Placement.rcNormalPosition;
    Values['RegValetFormBounds'] := IntToStr(R.Left) + ';' +
        IntToStr(R.Top) + ';' +
        IntToStr(R.Right - R.Left) + ';' +
        IntToStr(R.Bottom - R.Top) + ';';
  end;
end;

procedure TRegValet.SetIndexedValue(const Name: String; Index: Integer;
    const Value: String);
var
  L: TStringList;
begin
  L := GetIndexedListRef(Name);
  if Index >= L.Count then
    raise ERegValetError.Create('SetIndexedValue has invalid index');
  L[Index] := Value;
end;

procedure TRegValet.SetListDelim(const Value: String);
begin
  if Value = '' then
    raise ERegValetError.Create('ListDelim property may not be an ' +
        'empty string');
end;

procedure TRegValet.SetValue(const Name: String; const Value: String);
var
  I: Integer;
begin
  I := Names.IndexOf(Name);
  if I = -1 then
  begin
    Names.Add(Name);
    NonIndexedValues.Add(Value);
  end
  else
    NonIndexedValues[I] := Value;
end;

procedure TRegValet.StoreIndexedItem(const Name: String;
    Value: String);
var
  I: Integer;
  L: TStringList;
  S: String;
begin
  L := GetIndexedListRef(Name);
  AppendStr(Value, FListDelim);
  I := Pos(FListDelim, Value);
  while (I > 0) do
  begin
    S := Copy(Value, 1, (I - 1));
    if S <> '' then
      L.Add(S);
    Value := Copy(Value, (I + Length(FListDelim)), $7FFF);
    I := Pos(FListDelim, Value);
  end;
end;

procedure TRegValet.StoreItemFromReg(const Name: String;
    const Value: String);
begin
  if ItemIsIndexed(Name) then
    StoreIndexedItem(Name, Value)
  else
    StoreNonIndexedItem(Name, Value);
end;

procedure TRegValet.StoreNonIndexedItem(const Name: String;
    const Value: String);
var
  I: Integer;
begin
  I := Names.IndexOf(Name);
  if I <> -1 then
    NonIndexedValues[I] := Value
  else
  begin
    Names.Add(Name);
    NonIndexedValues.Add(Value);
  end;
end;

procedure TRegValet.WriteRegistry;
var
  I: Integer;
  Reg: TRegistry;

  procedure WriteIndexedItems(const Name: String; L: TStringList);
  var
    I: Integer;
    S: String;
  begin
    for I := 0 to (L.Count - 1) do
      S := S + L[I] + FListDelim;
    Reg.WriteString(Name, S);
  end;

begin
  Reg := TRegistry.Create;
  try
    Reg.RootKey := HKEY_CURRENT_USER;
    if Reg.OpenKey(FPath, False) then
      Reg.DeleteKey(FPath);
    Reg.OpenKey(FPath, True);
    for I := 0 to (Names.Count - 1) do
    begin
      if Names.Objects[I] <> nil then
        WriteIndexedItems(Names[I], TStringList(Names.Objects[I]))
      else
        Reg.WriteString(Names[I], NonIndexedValues[I]);
    end;
    Reg.CloseKey;
  finally
    Reg.Free;
  end;
end;

{ Register }

procedure Register;
begin
  RegisterComponents('Samples', [TRegValet]);
end;

end.
